from datetime import datetime, timedelta
from chaosmonkey.planners.planner import Planner


class TwoExecutors(Planner):
    """
    Schedule two executors
    """

    ref = "two_executors:TwoExecutors"
    schema = {}
    example = {}

    def plan(self, planner_config, attack_config):
        plan = self.add_plan(self.name)
        attack_date = datetime.now() + timedelta(hours=10)
        self._add_executor(attack_date, self.name, attack_config, plan.id)
        attack_date = attack_date + timedelta(hours=1)
        self._add_executor(attack_date, self.name, attack_config, plan.id)

    @staticmethod
    def to_dict():
        return Planner._to_dict(TwoExecutors.ref, TwoExecutors.schema, TwoExecutors.example)



from chaosmonkey.attacks.attack import Attack


class Attack2(Attack):

    schema = {}
    example = {}
    ref = "attack2:Attack2"

    def __init__(self, attack_config):
        super(Attack2, self).__init__(attack_config)

    def run(self):
        pass

    @staticmethod
    def to_dict():
        return Attack._to_dict(
            Attack2.ref,
            Attack2.schema,
            Attack2.example
        )



"""
ChaosMonkey Engine
"""
import os
import errno
from pytz import timezone

from chaosmonkey.api.app import flask_app
from chaosmonkey.engine.cme_manager import manager
from chaosmonkey.engine.scheduler import scheduler
from chaosmonkey.attacks.attack import Attack
from chaosmonkey.dal.cme_sqlalchemy_store import CMESQLAlchemyStore
from chaosmonkey.dal.database import db
from chaosmonkey.modules.module_store import ModulesStore
from chaosmonkey.planners.planner import Planner


def configure_engine(database_uri, attacks_folder, planners_folder, cme_timezone):
    """
    Create a Flask App and all the configuration needed to run the CMEEngine

    * Init and configure the SQLAlchemy store (create db and tables if don't exists)
    * Init ModuleStores (attacks and planners)
    * Configure the timezone and jobstore for the scheduler
    * Configure the CMEManager

    TODO:
        The scheduler start is not made until the first request is made. This is due to
        the way the SQLAlchemy store is created, because it needs the app.context to work
        properly

    :param database_uri:    SQLAlchemy SQLALCHEMY_DATABASE_URI
    :param attacks_folder:  folder to load the attacks modules
    :param planners_folder: folder to load the planners modules
    :param cme_timezone:    timezone to set in the scheduler
    """

    # configure and init FlaskSQLAlchemy
    if database_uri != ":memory:":
        # check for memory database for tests if not, make sure the path exists and create it
        make_sure_path_exists(os.path.dirname(database_uri))

    flask_app.config["SQLALCHEMY_DATABASE_URI"] = "sqlite:///%s" % database_uri
    flask_app.config["SQLALCHEMY_TRACK_MODIFICATIONS"] = False
    db.init_app(flask_app)
    with flask_app.app_context():
        db.create_all()
        db.app = flask_app

    # init stores
    sql_store = CMESQLAlchemyStore()
    planners_store = ModulesStore(Planner)
    attacks_store = ModulesStore(Attack)

    # configure the scheduler
    tz = timezone(cme_timezone)
    jobstores = {
        "default": sql_store
    }
    scheduler.configure(jobstores=jobstores, timezone=tz)

    # configure module stores
    attacks_store.load(attacks_folder)
    planners_store.load(planners_folder)

    # configure CMEManager
    manager.configure(scheduler, sql_store, planners_store, attacks_store)


# Start the scheduler in the first request
@flask_app.before_first_request
def start_scheduler():
    if not scheduler.running:
        scheduler.start()


def shutdown_engine():
    """
    Shutdown the scheduler
    """
    if scheduler.running:
        scheduler.shutdown()


def make_sure_path_exists(path):
    """
    Make sure a path exists and create it if don't

    :param path: string path to check
    """
    try:
        os.makedirs(path)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise



"""
CMESQLAlchemyStore replaces the default SQLAlchemyJobstore from
`apscheduler <http://apscheduler.readthedocs.io/>`_

It controls the persistence layer.
"""
import logging

from apscheduler.job import Job
from apscheduler.jobstores.base import BaseJobStore, JobLookupError
from sqlalchemy import text
from chaosmonkey.engine.cme_manager import manager
from chaosmonkey.dal.executor_model import Executor
from chaosmonkey.dal.plan_model import Plan
from chaosmonkey.dal.database import db

try:
    import cPickle as pickle
except ImportError:  # pragma: nocover
    import pickle


class CMESQLAlchemyStore(BaseJobStore):
    """
    CMESQLAlchemyStore

    Manage persistence for `apscheduler <http://apscheduler.readthedocs.io/>`_
    and CMEEngine.

    This class is used by the apscheduler, all overridden methods should
    return appscheduler.job.Job objects.

    The store handles 2 types of models: plans and executors.
    Internally apscheduler names the executors as jobs.

    TODO: executors are marked as executed when they are processed
        (no mather if they fail or succeed. We need to handle execution errors
        in order to know whats going on with the execution and what was its result.)

    * Plans: :meth:`chaosmonkey.dal.plan_model.Plan`
    * Executors: :meth:`chaosmonkey.dal.executor_model.Executor`
    """

    def __init__(self, pickle_protocol=pickle.HIGHEST_PROTOCOL):
        super(CMESQLAlchemyStore, self).__init__()  # pylint: disable=no-member
        self.pickle_protocol = pickle_protocol
        self.log = logging.getLogger(__name__)

    def start(self, scheduler, alias):
        """
        Start the SQLAlchemy engine
        """
        super(CMESQLAlchemyStore, self).start(scheduler, alias)

    def lookup_job(self, job_id):
        job = Executor.query.get(job_id)
        if not job:
            raise JobLookupError("job with id %s not found" % job_id)
        else:
            return self._reconstitute_job(job.job_state) if job.job_state else None

    def get_due_jobs(self, now):
        return self._get_jobs(Executor.next_run_time <= now)

    def get_next_run_time(self):
        job = Executor.query.order_by(Executor.next_run_time).limit(1).all()
        next_run_time = None
        if len(job) > 0:
            next_run_time = manager.scheduler.timezone.localize(job[0].next_run_time)
        return next_run_time

    def get_all_jobs(self):
        jobs = self._get_jobs()
        self._fix_paused_jobs_sorting(jobs)
        return jobs

    def add_job(self, job):
        job_state = pickle.dumps(job.__getstate__(), self.pickle_protocol)
        job = Executor(job.id, job.next_run_time, job.kwargs.get("plan_id"), job_state)
        db.session.add(job)
        db.session.commit()
        self._check_plan_executed(job.id)

    def update_job(self, job):
        job_model = Executor.query.get(job.id)
        job_model.next_run_time = job.next_run_time
        job_model.job_state = pickle.dumps(job.__getstate__(), self.pickle_protocol)
        db.session.commit()

    def remove_job(self, job_id):
        """
        Instead of deleting a job when its executed or it has failed, check it as executed.

        TODO: delete the executor and save to a historic table the executed attacks with its
        logs and results.
        """
        job_model = Executor.query.get(job_id)
        if job_model is None:
            raise JobLookupError(job_id)

        job_model.executed = True
        db.session.commit()

        self._check_plan_executed(job_id)

    def real_remove_job(self, job_id):
        self.log.debug('real remove job %s', job_id)
        job_model = Executor.query.get(job_id)
        if job_model is None:
            raise JobLookupError(job_id)

        db.session.delete(job_model)
        db.session.commit()

    def remove_all_jobs(self):
        db.session.query(Executor).delete()
        db.session.commit()

    def shutdown(self):
        pass

    def _reconstitute_job(self, job_state):
        job_state = pickle.loads(job_state)
        job_state['jobstore'] = self
        job = Job.__new__(Job)
        job.__setstate__(job_state)
        job._scheduler = self._scheduler  # pylint: disable=protected-access
        job._jobstore_alias = self._alias  # pylint: disable=protected-access
        return job

    def _get_jobs(self, *conditions):
        """
        Return only jobs with executed == 0. Because we are not deleting the executors we need
          to filter to ensure the apscheduler gets only pending jobs
        """
        job_list = []
        jobs = db.session.query(Executor).filter(Executor.executed == 0, *conditions)

        failed_job_ids = set()
        for job in jobs:
            try:
                job_list.append(self._reconstitute_job(job.job_state))
            except:  # pylint: disable=bare-except
                self.log.exception('Unable to restore job "%s" -- removing it', job.id)
                failed_job_ids.add(job.id)

        # Remove all the jobs we failed to restore
        if failed_job_ids:
            Executor.query.filter(Executor.id.in_(failed_job_ids))

        return job_list

    # Custom methods.
    # Methods defined bellow are only used by the CMEManager and must only return
    # db.Models (chaosmonkey.dal.*_model)

    def _check_plan_executed(self, executor_id):
        """
        Each time a executor is marked as executed check if all executors in the plan
        are executed and check the plan as executed as well

        :param executor_id: string
        :return:
        """
        executor = Executor.query.get(executor_id)
        if executor:
            plan = Plan.query.get(executor.plan_id)

            self.log.debug('check if all executors in plan %s are executed', executor.plan_id)
            # get all executor for the same plan that are not executed
            # pylint: disable=singleton-comparison
            executors_pending = Executor.query.\
                filter(Executor.executed == False, Executor.plan_id == executor.plan_id).all()
            if len(executors_pending) == 0:
                self.log.debug('all executors has been executed for the plan %s', executor.plan_id)
                # if no executors are pending to execute, mark the pan as executed

                plan.executed = True
            else:
                self.log.debug('there are executors pending in plan %s', executor.plan_id)
                plan.executed = False

            db.session.commit()

    def get_executor(self, executor_id):
        """
        Get an executor

        :param executor_id: string
        :return: List of Executor
        """
        self.log.debug('get executor %s', executor_id)
        return Executor.query.get(executor_id)

    def get_executors(self, executed=False):
        """
        Get a list of executors

        :return: List of Executor
        """
        self.log.debug('get all executors in store with executed %s', executed)
        return Executor.query.filter(Executor.executed == executed)\
            .order_by(Executor.next_run_time).all()

    def get_executors_for_plan(self, plan_id):
        """
        Get a list of executors related to a plan by its plan_id

        :param plan_id: string
        :return: List of Executor
        """
        self.log.debug('get executors for plan %s', plan_id)
        return db.session.query(Executor).filter(Executor.plan_id == plan_id)

    def add_plan(self, name):
        """
        Create a plan in the db.

        :param name: string
        :return: Plan created
        """
        self.log.debug('create plan %s', name)
        plan = Plan(name=name)
        db.session.add(plan)
        db.session.commit()
        return plan

    def get_plans(self, show_all=False):
        """
        Return a list of plans created on db. For each plan return the number of pending executors and the
        next_run_time of the first executor

        :return: List of Plans
        """

        query = 'SELECT ' \
                'cme_plans.id, name, created, COUNT(cme_executors.id) as executors_count, ' \
                ' MIN(next_run_time), cme_plans.executed ' \
                'FROM cme_plans ' \
                'LEFT JOIN cme_executors ON cme_plans.id == cme_executors.plan_id ' \

        if show_all is False:
            query += 'WHERE cme_plans.executed == 0 '

        query += 'GROUP BY cme_plans.id'
        self.log.debug('get plans query %s', query)
        sql = text(query)
        result = db.engine.execute(sql)
        plans = []
        for row in result:
            plan = Plan(
                _id=row[0],
                name=row[1],
                created=row[2],
                executors_count=row[3],
                next_execution=row[4],
                executed=row[5]
            )
            plans.append(plan)

        return plans

    def get_plan(self, plan_id):
        """
        Return a plan by its id

        :param plan_id: string
        :return: Plan
        """
        self.log.debug('get plan %s', plan_id)
        plan = Plan.query.get(plan_id)
        return plan

    def delete_plan(self, plan_id):
        """
        Delete a plan.

        All the executors related to the plan are deleted. (ON_DELETE constrain in db.Models)

        :param plan_id: string
        """
        self.log.debug('delete plan %s', plan_id)
        plan = Plan.query.get(plan_id)
        if plan:
            db.session.delete(plan)
            db.session.commit()
        else:
            raise PlanLookupError(plan_id)

    def __repr__(self):
        return '<%s>' % self.__class__.__name__


class PlanLookupError(KeyError):
    """Raised when the store cannot find a plan for update or removal."""

    def __init__(self, job_id):
        super(PlanLookupError, self).__init__(u'No plan by the id of %s was found' % job_id)



"""
Utils functions for blueprints methods
"""


def get_boolean(value):
    """
    to get boolean values from query strings

    :param value:
    :return: boolean
    """
    if isinstance(value, bool):
        return value

    if not value:
        return False
    value = value.lower()
    if value in ('false', 'no', 'off', 'n', '0',):
        return False
    if value in ('true', 'yes', 'on', 'y', '1',):
        return True

    return False



from chaosmonkey.attacks.attack import Attack


class RequiredSchema(Attack):

    ref = "required_schema:RequiredSchema"

    schema = {
        "type": "object",
        "properties": {
            "ref": {"type": "string"},
            "args": {
                "type": "object",
                "properties": {
                    "property1": {"type": "string"},
                    "property2": {"type": "string"},
                },
                "required": ["property1", "property2"]
            }
        }
    }

    example = {
        "ref": ref,
        "args": {
            "property1": "test1",
            "property2": "test2"
        }
    }

    def __init__(self, attack_config):
        super(RequiredSchema, self).__init__(attack_config)

    def run(self):
        pass

    @staticmethod
    def to_dict():
        return Attack._to_dict(
            RequiredSchema.ref,
            RequiredSchema.schema,
            RequiredSchema.example
        )



from flask import url_for
from flask_hal import Document
import test.planners.planner1 as planner1_module
import test.planners.planner2 as planner2_module


def test_empty_planners_store_return_hal(app, manager):
    url = url_for("planners.list_planners")

    planners_list = []
    manager.planners_store.set_modules(planners_list)

    with app.test_request_context(url):
        res = app.test_client().get(url)
        assert res.status_code == 200
        assert res.mimetype == "application/hal+json"
        assert res.json == Document(data={"planners": []}).to_dict()


def test_planners_list_return_hal(app, manager):
    url = url_for("planners.list_planners")

    module_list = [planner1_module, planner2_module]
    manager.planners_store.set_modules(module_list)

    with app.test_request_context(url):

        planner_list = [planner1_module.Planner1("planner1").to_dict(), planner2_module.Planner2("planner2").to_dict()]

        res = app.test_client().get(url)
        assert res.status_code == 200
        assert res.mimetype == "application/hal+json"
        assert res.json == Document(data={"planners": planner_list}).to_dict()



"""
**Base path**: /api/1/executors

Executors are scheduled jobs that are related with an attack, so in the given date
the job will execute the attack.
The only way to create executors is through :meth:`chaosmonkey.api.plans_blueprint`

Evey executor has 4 main properties:

1. **id**: unique identifier
2. **next_run_time**: The time and date that the executor is going to be executed
3. **plan_id**: id of the plan that created the executor
4. **executed**: if the executor has been executed

Example::

    {
        "id": "3b373155577b4d1bbc62216ffea013a4",
        "plan_id": "3ec72048cab04b76bdf2cfd4bc81cd1e",
        "next_run_time": "2017-01-25T10:12:1485339145",
        "executed": false
    }

"""
import arrow
from apscheduler.jobstores.base import JobLookupError
from apscheduler.triggers.date import DateTrigger
from flask import Blueprint, json, request
from flask_hal import Document

from chaosmonkey.api.api_errors import APIError
from chaosmonkey.api.request_validator import validate_payload
from chaosmonkey.engine.cme_manager import manager
from chaosmonkey.api.utils import get_boolean

executors = Blueprint("executors", __name__)

executor_trigger_schema = {
    "type": "object",
    "properties": {
        "type": {"type": "string"},
        "args": {
            "type": "object",
            "properties": {
                "date": {"type": "string"}
            },
            "required": ["date"]
        }
    },
    "required": ["type", "args"]
}


@executors.route("/", methods=["GET"])
def get_executors():
    """
    Get a list of scheduled executors

    Example response::

        {
            "executors":[
                {
                    "_links":{
                        "self":{
                            "href":"/api/1/executors/3b373155577b4d1bbc62216ffea013a4"
                        },
                        "update":{
                            "href":"/api/1/executors/3b373155577b4d1bbc62216ffea013a4"
                        },
                        "delete":{
                            "href":"/api/1/executors/3b373155577b4d1bbc62216ffea013a4"
                        }
                    },
                    "id":"3b373155577b4d1bbc62216ffea013a4",
                    "plan_id":"3ec72048cab04b76bdf2cfd4bc81cd1e",
                    "next_run_time":"2017-01-25T10:12:1485339145",
                    "executed": false
                }
            ]
        }

    :param: executed. Control when to show all executors (true) or only not executed (false). Defaults to false

    :return: :meth:`flask_hal.document`
    """
    executed_query = request.args.get("executed", False)
    executed = get_boolean(executed_query)
    executors_list = [executor.to_dict() for executor in manager.get_executors(executed=executed)]
    return Document(embedded={"executors": executors_list})


@executors.route("/<string:executor_id>", methods=['PUT'])
def put_executor(executor_id):
    """
    Update a executor to change its date. To provide a new date use the format
    in the example bellow.
    The format is used to create a
    `DateTrigger <https://github.com/agronholm/apscheduler/blob/master/apscheduler/triggers/date.py>`_
    from the apscheduler.

    TODO: create more `Triggers <https://github.com/agronholm/apscheduler/blob/master/apscheduler/triggers>`_

    Example request::

        PUT /api/1/executors/3b373155577b4d1bbc62216ffea013a4
        Body:
            {
              "type" : "date",
              "args" : {
                "date": "2017-10-23T19:19"
              }
            }

    Example response::

        {
          "id": "3b373155577b4d1bbc62216ffea013a4",
          "plan_id": "3ec72048cab04b76bdf2cfd4bc81cd1e",
          "next_run_time": "2017-10-23T19:19:1508786354",
          "executed": false,
          "_links": {
            "self": {
              "href": "/api/1/executors/3b373155577b4d1bbc62216ffea013a4"
            },
            "update":{
              "href":"/api/1/executors/3b373155577b4d1bbc62216ffea013a4"
            },
            "delete":{
              "href":"/api/1/executors/3b373155577b4d1bbc62216ffea013a4"
            }
          }
        }

    :return: :meth:`flask_hal.document`
    """
    assert validate_payload(request, executor_trigger_schema)
    body = request.get_json()

    trigger = dict_to_trigger(body)

    try:
        executor = manager.update_executor_trigger(executor_id, trigger)
    except JobLookupError:
        return json.jsonify({"msg": "executor not found " + executor_id}), 404
    else:
        return Document(data=executor.to_dict())


@executors.route("/<string:executor_id>", methods=['DELETE'])
def delete_executor(executor_id):
    """
    Delete an executor

    Example request::

        DEL /api/1/executors/6890192d8b6c40e5af16f13aa036c7dc
    """
    manager.remove_executor(executor_id)
    return json.jsonify({
        "msg": "Executor %s succesfully deleted" % executor_id
    })

########
# UTILS
#######


def trigger_to_dict(trigger):
    """
    Returns a dict version of the trigger
    """
    if isinstance(trigger, DateTrigger):
        trigger_dict = {}
        trigger_dict["type"] = "date"
        trigger_dict["args"] = {"date": arrow.get(trigger.run_date).isoformat()}
        return trigger_dict


def dict_to_trigger(trigger_dict):
    """
    Returns a trigger version of the trigger json
    """
    if trigger_dict["type"] == "date":
        if "args" not in trigger_dict:
            raise APIError("Missing trigger args")
        if "date" not in trigger_dict["args"]:
            raise APIError("Invalid trigger args")

        try:
            trigger = DateTrigger(run_date=arrow.get(trigger_dict["args"]["date"]).datetime)
        except arrow.parser.ParserError:
            raise APIError("Invalid date format")

        return trigger
    else:
        raise APIError("Invalid trigger type")



from flask import url_for
from flask_hal import Document
import test.attacks.attack1 as attack1_module
import test.attacks.attack2 as attack2_module


def test_empty_attack_store_return_hal(app, manager):
    url = url_for("attacks.list_attacks")

    attack_list = []
    manager.attacks_store.set_modules(attack_list)

    with app.test_request_context(url):
        res = app.test_client().get(url)
        assert res.status_code == 200
        assert res.mimetype == "application/hal+json"
        assert res.json == Document(data={"attacks": attack_list}).to_dict()


def test_attack_list_return_hal(app, manager):
    url = url_for("attacks.list_attacks")

    module_list = [attack1_module, attack2_module]
    manager.attacks_store.set_modules(module_list)

    with app.test_request_context(url):

        attack_list = [attack1_module.Attack1(None).to_dict(), attack2_module.Attack2(None).to_dict()]

        res = app.test_client().get(url)
        assert res.status_code == 200
        assert res.mimetype == "application/hal+json"
        assert res.json == Document(data={"attacks": attack_list}).to_dict()



"""
CME Manager

Control layer for CME Engine
"""
import logging

from jsonschema import validate, ValidationError
from chaosmonkey.dal.executor_model import Executor
from chaosmonkey.api.api_errors import APIError
from chaosmonkey.modules.module_store import ModuleLookupError


class CMEManager:
    """
    CMEManager is the manager responsible of communicating the API with the backend.

    It manages:

    * scheduler: BackgroundScheduler from appscheduler lib. The scheduler run executors
    * sql_store: SQLAlchemy store. The persistence layer
    * planners_store: ModuleStore that load and manages available planners.
    * attacks_store: ModuleStore that load and manage available attacks.

    Methods to interact with Executors and Plans always returns the db.Models (chaosmonkey.dal.*_model)

    Methods that interact with Attacks always returns Attacks objects (chaosmonkey.attacks.attack)

    Methods that interact with Planners always returns Planners objects (chaosmonkey.planners.planner)
    """
    def __init__(self):
        self._scheduler = None
        self._sql_store = None
        self._planners_store = None
        self._attacks_store = None
        self.log = logging.getLogger(__name__)

    def configure(self, scheduler, sql_store, planners_store, attacks_store):
        """
        Configure the manager

        :param scheduler:       apscheduler.schedulers.background.BackgroundScheduler
        :param sql_store:       SQLAlchemy
        :param planners_store:  chaosmonkey.modules.ModuleStore
        :param attacks_store:   chaosmonkey.modules.ModuleStore
        :return:
        """
        self._scheduler = scheduler
        self._sql_store = sql_store
        self._planners_store = planners_store
        self._attacks_store = attacks_store
        self.log.debug('CMEManager configured')

    @property
    def scheduler(self):
        """ Scheduler property """
        return self._scheduler

    @property
    def attacks_store(self):
        """ Attacks store property """
        return self._attacks_store

    @property
    def planners_store(self):
        """ Planners store property """
        return self._planners_store

    @property
    def sql_store(self):
        """ SQLstore property """
        return self._sql_store

    def get_executors(self, executed=False):
        """
        Return a list of Executor objects created in DB
        :return: chaosmonkey.dal.executor.Executor list
        """
        return self._sql_store.get_executors(executed=executed)

    def get_executor(self, executor_id):
        """
        Return an Executor object with the given id
        :return: chaosmonkey.dal.executor.Executor
        """
        return self._sql_store.get_executor(executor_id)

    def get_executors_for_plan(self, plan_id):
        """
        Return a list of Executors for a given plan id
        :return: chaosmonkey.dal.executor.Executor
        """
        return self.sql_store.get_executors_for_plan(plan_id)

    def update_executor_trigger(self, executor_id, trigger):
        """
        Update an executor trigger to change is due date

        :param executor_id: executor id
        :param trigger: apscheduler.triggers.BaseTrigger instance
        :return:        chaosmonkey.dal.executor.Executor
        """
        job = self._scheduler.reschedule_job(job_id=executor_id, trigger=trigger)
        return self._job_to_executor(job)

    def remove_executor(self, executor_id):
        """
        Removes an executor by his ID
        :param executor_id:
        :return:
        """
        self._scheduler.remove_job(job_id=executor_id)
        self.sql_store.real_remove_job(job_id=executor_id)

    def add_executor(self, date, name, attack_config, plan_id):
        """
        Adds a new executor to the scheduler

        :param date:            Datetime to execute the job
        :param name:            Executor name
        :param attack_config:   Attack config. Dict to be passed to the executor on execution time
        :param plan_id:         Referenced plan id
        :return:                chaosmonkey.dal.executor.Executor
        """
        from chaosmonkey.attacks.executor import execute
        self.log.debug('add scheduled job %s at %s', name, date)
        job = self._scheduler.add_job(
            execute,
            name=name,
            kwargs={
                "attack_config": attack_config,
                "plan_id": plan_id
            },
            trigger='date',
            run_date=date
        )

        return self._job_to_executor(job)

    def get_attack_list(self):
        """
        Return a list with all attacks loaded in the self._attacks_store

        :return: chaosmonkey.attacks.attack.Attack list
        """
        attacks_list = [self._attacks_store.get(attack_ref).to_dict() for attack_ref in self.attacks_store.list()]
        return attacks_list

    def get_planner_list(self):
        """
        Return a list with all planners loaded in the self._planners_store

        :return: chaosmonkey.planners.planner.Planner list
        """
        planner_list = [self._planners_store.get(planner_ref).to_dict() for planner_ref in self.planners_store.list()]
        return planner_list

    def execute_plan(self, name, planner_config, attack_config):
        """
        Execute a plan with a planner and executor config to create executors based on the configs

        It also validates the planner and executor config against the modules

        :param name:                Plan name
        :param planner_config:      Dict with planner config
        :param attack_config:       Dict with attack config
        """
        try:
            planner_class = self._planners_store.get(planner_config.get("ref"))
            attack_class = self._attacks_store.get(attack_config.get("ref"))
        except ModuleLookupError as e:
            raise APIError("invalid planner %s" % e.message)

        # Validate both executor and planner configs
        try:
            validate(planner_config.get("args"), planner_class.schema)
            validate(attack_config, attack_class.schema)
        except ValidationError as e:
            raise APIError("invalid payload %s" % e.message)

        planner = planner_class(name)
        planner.plan(planner_config, attack_config)

    def add_plan(self, name):
        """
        Creates a new plan in the sqlStore

        :param name:    Plan name
        :return:        chaosmonkey.dal.plan.Plan
        """
        return self._sql_store.add_plan(name)

    def get_plan(self, plan_id):
        """
        Returns a plans

        :return: chaosmonkey.dal.plan.Plan
        """
        return self._sql_store.get_plan(plan_id)

    def get_plans(self, show_all=None):
        """
        Returns a list with al plans in the sqlStore

        :return: List of chaosmonkey.dal.plan.Plan
        """
        return self._sql_store.get_plans(show_all=show_all)

    def delete_plan(self, plan_id):
        """
        Delete a plan (and all his associated executors) from the sqlStore

        :param plan_id:     String plan Id
        :return:
        """
        self._sql_store.delete_plan(plan_id)

    @staticmethod
    def _job_to_executor(job):
        """
        Converts a apscheduler.job.Job to chaosmonkey.dal.executor.Executor

        :param job:
        :return:
        """
        return Executor(
            job.id,
            job.next_run_time,
            job.kwargs.get("plan_id")
        )


manager = CMEManager()



import logging
from flask_cors import CORS
from flask import Flask, json
from flask_hal import HAL, HALResponse
from chaosmonkey.api.attacks_blueprint import attacks
from chaosmonkey.api.executors_blueprint import executors
from chaosmonkey.api.planners_blueprint import planners
from chaosmonkey.api.plans_blueprint import plans
from chaosmonkey.api.api_errors import APIError

log = logging.getLogger(__name__)

# Create FlaskApp
flask_app = Flask("cm_api")
HAL(flask_app, HALResponse)
CORS(flask_app)

# Register blueprints
root = "/api/"
prev1 = root + "1"
flask_app.register_blueprint(executors, url_prefix=prev1 + "/executors")
flask_app.register_blueprint(plans, url_prefix=prev1 + "/plans")
flask_app.register_blueprint(attacks, url_prefix=prev1 + "/attacks")
flask_app.register_blueprint(planners, url_prefix=prev1 + "/planners")

log.debug('Registered blueprints at %s', prev1)


# Register error handler for custom APIError exception
@flask_app.errorhandler(APIError)
def handle_invalid_usage(error):
    """
    Handler for APIErrors thrown by API endpoints
    """
    log.info('%d %s', error.status_code, error.message)
    response = json.jsonify(error.to_dict())
    response.status_code = error.status_code
    return response



from chaosmonkey.planners.planner import Planner


class Planner2(Planner):
    """
    Schedule a single job for a specific date
    """

    ref = "planner1:Planner2"
    schema = {}
    example = {}

    def plan(self, planner_config, attack_config):
        pass

    @staticmethod
    def to_dict():
        return Planner._to_dict(Planner2.ref, Planner2.schema, Planner2.example)



"""
**Base path**: /api/1/planners

Planners are python modules (located in /planners folder). Planners are responsible of create
:meth:`executors <chaosmonkey.api.executors_blueprint>`.

Planners has three main properties represented in the API:

1. **example**:  a JSON example for the planner
2. **ref**: its unique identifier. module_name:PlannerClass
3. **schema**: json schema that validates the planner

"""
from flask import Blueprint
from flask_hal import Document
from chaosmonkey.engine.cme_manager import manager

planners = Blueprint("planners", __name__)


@planners.route("/", methods=["GET"])
def list_planners():
    """
    Return a list with the available planners and its configuration.

    Example response::

        {
            "_links": {
                "self": {
                    "href": "/api/1/planners/"
                }
            },
            "planners": [
                {
                    "example": {
                        "args": {
                            "times": 4,
                            "max_time": "15:00",
                            "min_time": "10:00"
                        },
                        "ref": "simple_planner:SimplePlanner"
                    },
                    "ref": "simple_planner:SimplePlanner",
                    "schema": {
                        "type": "object",
                        "properties": {
                            "args": {
                                "type": "object",
                                "properties": {
                                    "times": {
                                        "type": "number"
                                    },
                                    "max_time": {
                                        "type": "string"
                                    },
                                    "min_time": {
                                        "type": "string"
                                    }
                                }
                            },
                            "ref": {
                                "type": "string"
                            }
                        }
                    }
                }
            ]
        }

    :return: :meth:`flask_hal.document`
    """
    planners_list = manager.get_planner_list()
    return Document(data={"planners": planners_list})



"""
API Tests
"""
import arrow
from apscheduler.triggers.date import DateTrigger
from flask import url_for
from flask_hal import Document
from datetime import datetime, timedelta

from chaosmonkey.api.executors_blueprint import trigger_to_dict, dict_to_trigger


def test_executors_list_return_empty_array(app):
    url = url_for("executors.get_executors")

    with app.test_request_context(url):

        res = app.test_client().get(url)
        assert res.status_code == 200
        assert res.mimetype == "application/hal+json"
        assert res.json == Document(embedded={"executors": []}).to_dict()


def test_get_executors(app, manager, plan):
    """" Test that the endpoint returns an array of executors """
    url = url_for("executors.get_executors")

    run_time = datetime.now() + timedelta(hours=10)
    executor = manager.add_executor(run_time, "executor name", {}, plan.id)

    with app.test_request_context(url):
        res = app.test_client().get(url)
        assert res.status_code == 200
        assert res.mimetype == "application/hal+json"
        assert res.json == Document(embedded={"executors": [executor.to_dict()]}).to_dict()


def test_put_executor_valid_body(app, manager, plan):
    # Add a executor to the datastore
    run_time = datetime.now() + timedelta(hours=10)
    executor = manager.add_executor(run_time, "executor name", {}, plan.id)

    # Make the request
    url = url_for("executors.put_executor", executor_id=executor.id)
    with app.test_request_context(url):
        data = '{"type": "date", "args": {"date": "%s"}}' % run_time.strftime('%Y-%m-%dT%H:%M:%s%z')
        res = app.test_client().put(url,
                                    content_type="application/json",
                                    data=data)

    assert res.status_code == 200
    assert res.mimetype == "application/hal+json"


def test_put_executor_missing_trigger_type(app):
    url = url_for("executors.put_executor", executor_id="123")
    run_time = datetime.now() + timedelta(hours=10)

    with app.test_request_context(url):
        data = '{"args":{"date":"%s"}}' % run_time.strftime('%Y-%m-%dT%H:%M:%s%z')
        res = app.test_client().put(url_for("executors.put_executor", executor_id="123"),
                                    content_type="application/json",
                                    data=data)

    assert res.status_code == 400
    assert res.mimetype == "application/json"
    assert res.json == {"msg": "invalid payload 'type' is a required property"}


def test_put_executor_invalid_trigger_type(app):
    url = url_for("executors.put_executor", executor_id="123")
    run_time = datetime.now() + timedelta(hours=10)

    with app.test_request_context(url):
        data = '{"type":"invalid", "args":{"date":"%s"}}' % run_time.strftime('%Y-%m-%dT%H:%M:%s%z')
        res = app.test_client().put(url_for("executors.put_executor", executor_id="123"),
                                    content_type="application/json",
                                    data=data)

    assert res.status_code == 400
    assert res.mimetype == "application/json"
    assert res.json == {"msg": "Invalid trigger type"}


def test_put_executor_missing_trigger_args(app):
    url = url_for("executors.put_executor", executor_id="123")

    with app.test_request_context(url):
        data = '{"type":"date"}'
        res = app.test_client().put(url_for("executors.put_executor", executor_id="123"),
                                    content_type="application/json",
                                    data=data)

    assert res.status_code == 400
    assert res.mimetype == "application/json"
    assert res.json == {"msg": "invalid payload 'args' is a required property"}


def test_put_executor_invalid_trigger_date(app):
    url = url_for("executors.put_executor", executor_id="123")

    with app.test_request_context(url):
        data = '{"type":"date","args":{"date":"invaliddate"}}'
        res = app.test_client().put(url_for("executors.put_executor", executor_id="123"),
                                    content_type="application/json",
                                    data=data)

    assert res.status_code == 400
    assert res.mimetype == "application/json"
    assert res.json == {"msg": "Invalid date format"}


def test_put_executor_valid_body_invalid_executorid(app):
    url = url_for("executors.put_executor", executor_id="123")
    run_time = datetime.now() + timedelta(hours=10)

    with app.test_request_context(url):
        data = '{"type": "date", "args": {"date":"%s"}}' % run_time.strftime('%Y-%m-%dT%H:%M:%s%z')
        res = app.test_client().put(url_for("executors.put_executor", executor_id="invalid"),
                                    content_type="application/json",
                                    data=data)

    assert res.status_code == 404
    assert res.mimetype == "application/json"
    assert res.json == {"msg": "executor not found invalid"}


def test_delete_executor(app, manager, plan):
    # Add a executor to the datastore
    run_time = datetime.now() + timedelta(hours=10)
    executor = manager.add_executor(run_time, "executor name", {}, plan.id)
    url = url_for("executors.delete_executor", executor_id=executor.id)
    with app.test_request_context(url):
        res = app.test_client().delete(url)

    assert res.status_code == 200
    assert res.mimetype == "application/json"
    assert res.json == {"msg": "Executor %s succesfully deleted" % executor.id}


#############
# UTILS TESTS
#############

def test_date_trigger_to_dict():
    my_dt = arrow.get("2016-06-21T15:30:12").datetime
    trigger = DateTrigger(run_date=my_dt)
    trigger_dict = {"type": "date", "args": {"date": "2016-06-21T15:30:12+00:00"}}
    assert trigger_dict == trigger_to_dict(trigger)


def test_json_to_date_trigger():
    my_dt = arrow.get("2016-06-21T15:30:12").datetime
    trigger_dict = {"type": "date", "args": {"date": "2016-06-21T15:30:12+00:00"}}
    trigger = DateTrigger(run_date=my_dt)
    assert trigger.run_date == dict_to_trigger(trigger_dict).run_date



from chaosmonkey.planners.planner import Planner


class Planner1(Planner):
    """
    Schedule a single job for a specific date
    """

    ref = "planner1:Planner1"
    schema = {}
    example = {}

    def plan(self, planner_config, attack_config):
        pass

    @staticmethod
    def to_dict():
        return Planner._to_dict(Planner1.ref, Planner1.schema, Planner1.example)



"""
This module contains an object that represent an API Error

Any APIError thrown in an endpoint is handled to return to the user
a proper json error with custom status code and message
"""


class APIError(Exception):
    """
    Represents an API Error

    :param message: message to be returned to the user
    :param status_code: response status code (defaults to 400)
    :param payload: custom payload to give extra info in the response

    Example:
            >>> raise APIError("Error detected", 500, {"extra": "extra_info"})

    """

    def __init__(self, message, status_code=None, payload=None):
        Exception.__init__(self)
        self.message = message
        self.status_code = status_code or 400
        self.payload = payload

    def to_dict(self):
        """Convert exception to dict"""
        dict_ = dict(self.payload or ())
        dict_['msg'] = self.message
        return dict_

    def __str__(self):
        return '[%d] %s' % (self.status_code, self.message)



'''
ModulesStore
Handle the dynamic load of modules outside the main package.

'''
import logging
import sys
import inspect
from os import listdir
from os.path import isfile, join, splitext, isdir


class ModulesStore:
    """
    Load and store modules from outside the app package
    """
    modules = []

    def __init__(self, klass):
        self.klass = klass
        self.modules = []
        self.log = logging.getLogger('%s.%s' % (__name__, klass.__name__))

    def load(self, path):
        """
        Loads all modules found in a given path.
        It adds the path to the sys.path and import all modules found

        :param path: path for lookup modules
        """

        self._validate_path(path)

        sys.path.insert(0, path)

        module_names = self._get_module_names(path)
        try:
            for name in module_names:
                self.log.debug('added module %s', name)
                module = __import__(name)
                self.modules.append(module)
        except ImportError:
            self.log.debug('error importing module %s', name)
            raise ValueError('Unable to import %s' % name)

    def set_modules(self, modules):
        self.modules = modules

    def get_modules(self):
        return self.modules

    def list(self):
        """
        List all loaded modules names. It will return a list withe the
        __name__ of each module
        :return:
        """
        module_names = []
        for module in self.get_modules():
            module_name = getattr(module, '__name__')
            for name, data in inspect.getmembers(module):
                if inspect.isclass(data) and self._has_klass_ancestor(data):
                    module_names.append(module_name + ':' + name)

        return module_names

    def add(self, module):
        self.modules.append(module)

    def remove(self, module_name):
        for module in self.modules:
            if module_name == getattr(module, '__name__'):
                self.modules.remove(module)

    def get(self, ref):
        """
        Given a str reference of a class in a module (eg. ModuleName.ClassName)
        return the class so it can be instantiated

        The module with the class must be loaded first using load or add method.

        :param ref: str representation of a class in a module
        :return: class ready for instantation
        """
        return self._ref_to_obj(ref)

    def _has_klass_ancestor(self, cls):
        if hasattr(cls, '__bases__') and self.klass in cls.__bases__:
            return True
        return False

    @staticmethod
    def _validate_path(path):
        '''
        Check for a valid path
        '''
        if not isdir(path):
            raise ValueError('invalid path ' + path)

    @staticmethod
    def _get_module_names(path):
        '''Inspect a path and return all file names without extension'''
        names = []
        for f in listdir(path):
            if isfile(join(path, f)):
                names.append(splitext(f)[0])
        return names

    def _ref_to_obj(self, ref):
        """
        Return the object pointed to by ``ref``.
        :type ref: str
        """
        if not isinstance(ref, str):
            raise TypeError('References must be strings')
        if ':' not in ref:
            raise ValueError('Invalid reference')

        module_name, class_name = ref.split(':', 1)
        for module in self.modules:
            attr_name = getattr(module, '__name__')
            if attr_name == module_name:
                try:
                    self.log.debug(attr_name + ' found on module store!')
                    return getattr(module, class_name)
                except Exception:
                    raise ModuleLookupError('Error resolving reference %s: not found in modules' % ref)


class ModuleLookupError(Exception):
    """
    Represent a Module lookup error when a module is not found by its ref
    """

    def __init__(self, message):
        Exception.__init__(self)
        self.message = message

    def __str__(self):
        return '%s' % self.message



import pytest
from chaosmonkey.engine.app import configure_engine, shutdown_engine
from chaosmonkey.engine.cme_manager import manager as cme_manager
from chaosmonkey.api.app import flask_app


def pytest_addoption(parser):
    """Options to configure attacks and planners root"""
    parser.addoption("--root", action="store", default="./",
                     help="src root folder")


@pytest.fixture(scope="session")
def root_src(request):
    return request.config.getoption("--root")


@pytest.yield_fixture(scope="session")
def app(root_src):
    """Crete app for test with context."""
    database_uri = ":memory:"
    attacks_folder = root_src + "attacks"
    planners_folder = root_src + "planners"
    cme_timezone = "Europe/Madrid"

    configure_engine(database_uri, attacks_folder, planners_folder, cme_timezone)
    cme_manager.scheduler.start()
    yield flask_app
    tear_down()


def tear_down():
    shutdown_engine()


@pytest.fixture
def manager():
    return cme_manager


@pytest.fixture
def plan(manager):
    plan = manager.add_plan("plan name")
    yield plan
    manager.delete_plan(plan.id)



"""
**Base path**: /api/1/attacks

Attacks are python modules (located in /attacks folder) that are executed to perform actual attacks.

Each attack has three main properties represented in the API:

1. **example**: a JSON example for the attack. Use it as a template to call /plans endpoints
2. **ref**: its unique identifier. module_name:AttackClass
3. **schema**: json schema that validates the json representation for the attack

"""
from flask import Blueprint
from flask_hal import Document
from chaosmonkey.engine.cme_manager import manager

attacks = Blueprint("attacks", __name__)


@attacks.route("/", methods=["GET"])
def list_attacks():
    """
    Return a list with the available attacks and its configuration.

    Example::

        {
            "attacks": [
                {
                    "example": {
                        "args": {
                            "filters": {
                                "tag:Name": "playground-asg"
                            },
                            "region": "eu-west-1"
                        },
                        "ref": "terminate_ec2_instance:TerminateEC2Instance"
                    },
                    "ref": "terminate_ec2_instance:TerminateEC2Instance",
                    "schema": {
                        "type": "object",
                        "properties": {
                            "args": {
                                "type": "object",
                                "properties": {
                                    "filters": {
                                        "type": "object",
                                        "properties": {
                                            "tag:Name": {
                                                "type": "string"
                                            }
                                        },
                                        "required": [
                                            "tag:Name"
                                        ]
                                    },
                                    "region": {
                                        "optional": true,
                                        "type": "string"
                                    }
                                },
                                "required": [
                                    "region",
                                    "filters"
                                ]
                            },
                            "ref": {
                                "type": "string"
                            }
                        }
                    }
                }
            ],
            "_links": {
                "self": {
                    "href": "/api/1/attacks/"
                }
            }
        }

    :return: :meth:`flask_hal.document`
    """
    attack_list = manager.get_attack_list()
    return Document(data={"attacks": attack_list})



from datetime import datetime
from uuid import uuid4
from sqlalchemy.orm import relationship
from flask import request
from flask_hal.document import BaseDocument, link
from chaosmonkey.dal.database import db
from chaosmonkey.dal.executor_model import Executor


class Plan(db.Model):
    """
    Store information about the plan and its executors.

    This model is only used by the cme.
    """
    __tablename__ = 'cme_plans'

    id = db.Column(db.String(80), primary_key=True)  #: unique identifier
    name = db.Column(db.String(200), unique=False)  #: plan name
    created = db.Column(db.DateTime, unique=True)  #: creation datetime
    executed = db.Column(db.Boolean)  #: if all the executors in the plan has been executed

    next_execution = None  #: DateTime for the next executor execution time
    executors_count = None  #: number of pending executors

    jobs = relationship(Executor, cascade='all, delete, delete-orphan')

    # pylint: disable=too-many-arguments
    def __init__(self, _id=None, name=None, created=None, next_execution=None, executors_count=0, executed=False):
        self.id = _id or uuid4().hex
        self.name = name
        if created:
            self.created = datetime.strptime(created, '%Y-%m-%d %H:%M:%S.%f')
        else:
            self.created = datetime.utcnow()
        self.executed = executed
        self.next_execution = next_execution
        self.executors_count = executors_count

    def to_dict(self):
        """
        Returns a :meth:`flask_hal.document` representation for the Executor

        :return: :meth:`chaosmonkey.dal.plan_model.HalPlan`
        """

        return HalPlan(data={
            'id': self.id,
            'name': self.name,
            'created': self.created.strftime('%Y-%m-%dT%H:%M:%s'),
            'executors_count': self.executors_count,
            'next_execution': self.next_execution,
            'executed': self.executed
        }).to_dict()

    def __repr__(self):
        return '<Plan %r>' % self.name


class HalPlan(BaseDocument):
    """
    Class to represent a Plan as a :meth:`flask_hal.document`
    """
    def __init__(self, data=None, links=None, embedded=None):
        super(HalPlan, self).__init__(data, links, embedded)

        self.links.append(link.Link("self", request.path + data["id"]))
        self.links.append(link.Link("update", request.path + data["id"]))
        self.links.append(link.Link("delete", request.path + data["id"]))



import os
import logging
import requests
from libcloud.compute.types import Provider
from libcloud.compute.providers import get_driver


class EC2DriverFactory:
    """
    Driver factory to get a libcloud driver with appropriate credentials for AWS provider
    You can provide credentials by either:

    * Setting AWS_IAM_ROLE in env variables
    * Setting AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY in env variables

    You can provide the region to connect by either:

    * Provide it at instantiation time
    * Setting AWS_DEFAULT_REGION in env variables
    """

    def __init__(self, region=None):
        """
        Initialize an EC2 driver factory for a certain AWS region

        :param region: The AWS region to operate within
        :type region: string
        """
        self.log = logging.getLogger(__name__)
        self.IAM_METADATA_URL = "http://169.254.169.254/latest/meta-data/iam/security-credentials"

        # First check if AWS_IAM_ROLE is defined
        aws_iam_role = os.environ.get("AWS_IAM_ROLE", None)
        if aws_iam_role is not None:
            # Get credentials from IAM role
            self.aws_ak, self.aws_sk, self.token = \
                self._get_aws_credentials_from_iam_role(aws_iam_role)
        else:
            # Get credentials from environment variables
            self.log.debug('getting AWS credentials from environment variables AWS_ACCESS_KEY_ID, '
                           'AWS_SECRET_ACCESS_KEY and AWS_DEFAULT_REGION')
            self.aws_ak = os.environ.get('AWS_ACCESS_KEY_ID')
            self.aws_sk = os.environ.get('AWS_SECRET_ACCESS_KEY')
            self.region = region if region is not None else os.environ.get("AWS_DEFAULT_REGION")

    def get_driver(self):
        """
        Return a Libcloud driver for AWS EC2 Provider

        :return: Compute driver
        :type driver: Libcloud compute driver
        """
        return (get_driver(Provider.EC2))(self.aws_ak, self.aws_sk, region=self.region)

    def _get_aws_credentials_from_iam_role(self, role):
        """
        With a valid IAM_ROLE make a request to the AWS metadata server to
        get temporary credentials for the role

        :param role: The IAM role to use
        :type role: string
        """
        url = "/".join((self.IAM_METADATA_URL, role))
        self.log.debug("get AWS credentials from IAM_METADATA (%s)", url)
        response = requests.get(url)
        response.raise_for_status()
        resp_json = response.json()
        aws_ak = resp_json.get("AccessKeyId")
        aws_sk = resp_json.get("SecretAccessKey")
        aws_token = resp_json.get("Token")
        return aws_ak, aws_sk, aws_token



"""
:meth:`apscheduler.schedulers.background.BackgroundScheduler` used by the CME.

The scheduler is responsible of storing executors and execute them in the given datatime.
"""

from apscheduler.schedulers.background import BackgroundScheduler

scheduler = BackgroundScheduler()



from unittest.mock import patch
import sys
import pytest
from chaosmonkey.modules.module_store import ModulesStore
from chaosmonkey.attacks.attack import Attack
import chaosmonkey.attacks.attack as module_attack


class AttackMock(Attack):
    @staticmethod
    def to_dict():
        pass

    def run(self):
        pass


def test_load_raise_exception_for_invalid_path():
    modules = ModulesStore(Attack)
    with pytest.raises(ValueError):
        modules.load("")


@patch.object(ModulesStore, "_get_module_names")
def test_load_path_with_no_modules(get_module_names_mock):
    get_module_names_mock.return_value = ["InvalidModule"]
    modules = ModulesStore(Attack)
    with pytest.raises(ValueError):
        modules.load("/tmp")


@patch.object(ModulesStore, "_get_module_names")
@patch.object(ModulesStore, "_validate_path")
def test_load_path_with_valid_modules(validate_path_mock, get_module_names_mock):
    module_list = ["sys", "os"]
    get_module_names_mock.return_value = module_list
    validate_path_mock.return_value = True
    modules_store = ModulesStore(Attack)
    modules_store.load("/tmp")
    assert len(modules_store.modules) == len(module_list)


def test_get_module_names():
    with patch("chaosmonkey.modules.module_store.listdir") as listdir_mock, \
         patch("chaosmonkey.modules.module_store.isfile") as isfile_mock:

        listdir_mock.return_value = ["test.py", "tost.py"]
        isfile_mock.return_value = True

        modules_store = ModulesStore(Attack)
        module_names = modules_store._get_module_names("")
        assert len(module_names) == 2
        assert module_names == ["test", "tost"]


def test_ref_to_obj():
    modules_store = ModulesStore(Attack)
    modules_store.add(module_attack)
    obj = modules_store._ref_to_obj("chaosmonkey.attacks.attack:Attack")
    assert obj is Attack


@patch.object(ModulesStore, "get_modules")
def test_list(modules_mock):
    sys.path.insert(0, "attacks")
    module = __import__("terminate_ec2_instance")
    modules_store = ModulesStore(Attack)
    modules_mock.return_value = [module]

    assert len(modules_store.list()) == 1


def test_add():
    sys.path.insert(0, "attacks")
    module = __import__("terminate_ec2_instance")
    modules_store = ModulesStore(Attack)
    modules_store.add(module)
    assert len(modules_store.list()) == 1


def test_remove():
    sys.path.insert(0, "attacks")
    module = __import__("terminate_ec2_instance")
    modules_store = ModulesStore(Attack)
    modules_store.add(module)
    modules_store.remove("terminate_ec2_instance")
    assert len(modules_store.list()) == 0



from chaosmonkey.api.api_errors import APIError
from jsonschema import validate, ValidationError


def validate_payload(request, schema):
    """
    validates a request payload against a json schema

    :param request: request received with valid json body
    :param schema:  schema to validate the request payload

    :return: True
    :raises: :meth:`chaosmonkey.api.api_errors`
    """
    try:
        json = request.get_json()
        validate(json, schema)
    except ValidationError as e:
        raise APIError("invalid payload %s" % e.message)
    except Exception:
        raise APIError("payload must be a valid json")
    else:
        return True



"""
Base class for planners

Every planner must extend Planner class
"""
from chaosmonkey.engine.cme_manager import manager


class Planner:
    """
    Planner interface
    Planners are responsible for scheduling jobs that executes attacks

    :param name:    plan name
    """

    ref = None  # must override
    schema = None  # must override
    example = None  # must override

    def __init__(self, name):
        self.name = name

    def plan(self, planner_config, attack_config):
        """
        Plan the jobs.
        This method should use the config to schedule jobs based on the
        configuration for the planner

        :param planner_config:    configuration related to the scheduler
        :param attack_config:   configuration related to the attack
        """
        raise NotImplementedError("Plans should implement this!")

    @staticmethod
    def add_plan(name):
        return manager.add_plan(name)

    @staticmethod
    def _add_executor(date, name, attack_config, plan_id):
        """
        Add a job to the global scheduler

        :param date: date to execute the job
        :param name: job name
        :param attack_config: configuration related to the attack
        """
        date_timezone = manager.scheduler.timezone.localize(date)
        manager.add_executor(date_timezone, name, attack_config, plan_id)

    @staticmethod
    def to_dict():
        raise NotImplementedError("Planners should implement this!")

    @staticmethod
    def _to_dict(ref, schema, example):
        return {
            "ref": ref,
            "schema": schema,
            "example": example,
        }



import logging
import chaosmonkey.engine.cme_manager as CMEManager

log = logging.getLogger(__name__)


def execute(attack_config=None, plan_id=None):
    """
    This func is executed for every job stored in the scheduler.
    Receive in kwargs all attack configuration used when creating
    the executor that indicates which attack and configuration should be
    used to do the actual attack.

    :param attack_config: **Dict** with attack configuration
    :param plan_id: **String** plan id for the plan containing the executor
    """
    attack_class = CMEManager.manager.attacks_store.get(attack_config.get('ref'))

    if attack_class is None:
        msg = '[PlanID %s] Attack ref %s not loaded in the store ' \
              % (plan_id, attack_config.get('ref'))
        log.debug(msg)
        raise ValueError(msg)

    attack_instance = attack_class(attack_config.get("args"))
    attack_instance.run()



from chaosmonkey.attacks.attack import Attack


class Attack1(Attack):

    schema = {}
    example = {}
    ref = "attack1:Attack1"

    def __init__(self, attack_config):
        super(Attack1, self).__init__(attack_config)

    def run(self):
        pass

    @staticmethod
    def to_dict():
        return Attack._to_dict(
            Attack1.ref,
            Attack1.schema,
            Attack1.example
        )



import logging
import signal
import click
from gevent.wsgi import WSGIServer
from chaosmonkey.engine.app import configure_engine, shutdown_engine
from chaosmonkey.api.app import flask_app


@click.command(name='chaos-monkey-engine')
@click.option("--port", "-p", default=5000, help="Port used to expose the CM API. Default 5000")
@click.option("--timezone", "-t", default="Europe/Madrid", help="Timezone to configure the scheduler. "
                                                                "Default 'Europe/Madrid'")
@click.option("--database-uri", "-d", required=True, help="SQLAlchemy database uri")
@click.option("--attacks-folder", "-a", required=True, help="Path to the folder where the attacks are stored")
@click.option("--planners-folder", "-p", required=True, help="Path to the folder where the planners are stored")
def cm(port, timezone, database_uri, attacks_folder, planners_folder):
    """
    Chaos Monkey Engine command line utility

    """
    log = logging.getLogger(__name__)
    configure_engine(database_uri, attacks_folder, planners_folder, timezone)

    log.info("Engine configured")
    log.debug("database: %s", database_uri)
    log.debug("attacks folder: %s", attacks_folder)
    log.debug("planners folder: %s", planners_folder)
    log.debug("timezone: %s", timezone)

    try:
        # Catch SIGTERM and convert it to a SystemExit
        signal.signal(signal.SIGTERM, sigterm_handler)
        log.info("Serving API at port %s", port)
        run_api_server(flask_app, port)
    except (KeyboardInterrupt, SystemExit):
        shutdown_engine()


def run_api_server(app, port=5000):
    """
    Runs a gevent server on a given port. Default port is 5000.

    :param app: WSGI-compliant app to be run
    :param port: port for the gevent server
    :type port: int
    :return: None
    """
    log = logging.getLogger('api')

    http_server = WSGIServer(('', port), app, log=log)
    http_server.serve_forever()


def sigterm_handler():
    raise SystemExit(1)



from datetime import datetime, timedelta
from flask import url_for, json
from flask_hal import Document
import test.attacks.attack1 as attack1_module
import test.planners.planner1 as planner1_module

valid_request_body = {
    "name": "Test Planner",
    "attack": {
        "ref": "test.attacks.attack1:Attack1",
        "args": {}
    },
    "planner": {
        "ref": "test.planners.planner1:Planner1",
        "args": {}
    }
}


def test_empty_plans_return_hal(app):
    url = url_for("plans.list_plans")

    with app.test_request_context(url):
        res = app.test_client().get(url)
        assert res.status_code == 200
        assert res.mimetype == "application/hal+json"
        assert res.json == Document(data={"plans": []}).to_dict()


def test_plan_list_return_hal(app, manager):
    url = url_for("plans.list_plans")

    plan = manager.add_plan("plan name")

    with app.test_request_context(url):

        plan_list = [plan.to_dict()]

        res = app.test_client().get(url)
        assert res.status_code == 200
        assert res.mimetype == "application/hal+json"
        assert res.json == Document(data={"plans": plan_list}).to_dict()


def test_plan_get_return_hal_with_executors(app, manager):
    plan = manager.add_plan("plan name")
    run_time = datetime.now() + timedelta(hours=10)
    executor = manager.add_executor(run_time, "executor name", {}, plan.id)

    url = url_for("plans.get_plan", plan_id=plan.id)

    with app.test_request_context(url):
        res = app.test_client().get(url)

        expected = Document(data=plan.to_dict(), embedded={"executors": [executor.to_dict()]}).to_dict()

        assert res.status_code == 200
        assert res.mimetype == "application/hal+json"
        assert res.json == expected


def test_plan_add_valid_body(app, manager):
    url = url_for("plans.add_plan")

    # add a planner and an attack
    manager.attacks_store.add(attack1_module)
    manager.planners_store.add(planner1_module)

    with app.test_request_context(url):

        res = app.test_client().post(url_for("plans.add_plan"),
                                     content_type="application/json",
                                     data=json.dumps(valid_request_body))

        assert res.status_code == 200
        assert res.mimetype == "application/json"
        assert res.json == {"msg": "ok"}



from chaosmonkey.planners.planner import Planner


class RequiredSchema(Planner):

    schema = {
        "type": "object",
        "properties": {
            "property1": {"type": "string"},
            "property2": {"type": "string"},
        },
        "required": ["property1", "property2"]
    }

    example = {
        "property1": "test1",
        "property2": "test2"
    }

    ref = "required_schema:RequiredSchema"

    def plan(self, planner_config, attack_config):
        pass

    @staticmethod
    def to_dict():
        return Planner._to_dict(RequiredSchema.ref, RequiredSchema.schema, RequiredSchema.example)



"""
This package contains the pieces to run the Engine

* Engine configuration (FlaskApp, Store, Modules...)
* CMEManager, used to communicate the Engine with the API and the Store
* Scheduler to schedule executors for later use
"""



"""
Chaos Monkey Engine main package
"""

# Configure default logging for entire app
import logging

logging.basicConfig(
    level=logging.DEBUG,
    format='%(asctime)s - %(levelname)s - %(name)s: %(message)s'
)

# different log level for apscheduler to avoid debug flood
logging.getLogger("apscheduler.scheduler").setLevel(logging.INFO)



from flask import request
from flask_hal.document import BaseDocument, link
from chaosmonkey.dal.database import db


class Executor(db.Model):
    """
    Executors are persistent representations of scheduled jobs.
    This model is shared between the cme and apscheduler.
    """

    __tablename__ = 'cme_executors'

    id = db.Column(db.String(80), primary_key=True)  #: unique identifier
    next_run_time = db.Column(db.DateTime, index=True)  #: DateTime for the executor to be executed
    job_state = db.Column(db.LargeBinary, nullable=False)  #: store the full state of the executor (with pickle)
    plan_id = db.Column(db.Integer, db.ForeignKey('cme_plans.id'))  #: plan id reference
    executed = db.Column(db.Boolean)  #: if the job was executed

    def __init__(self, job_id, next_run_time, plan_id, job_state=None):
        self.id = job_id
        self.next_run_time = next_run_time
        self.job_state = job_state
        self.plan_id = plan_id
        self.executed = False

    def to_dict(self):
        """
        Return a :meth:`flask_hal.document` representation for the Executor

        :return: :meth:`chaosmonkey.dal.executor_model.HalExecutor`
        """

        return HalExecutor(data={
            "id": self.id,
            "next_run_time": self.next_run_time.strftime('%Y-%m-%dT%H:%M:%s'),
            "plan_id": self.plan_id,
            "executed": self.executed
        }).to_dict()

    def __repr__(self):
        return '<Executor %r>' % self.id


class HalExecutor(BaseDocument):
    """
    Class to represent an Executor as a :meth:`flask_hal.document`
    """
    def __init__(self, data=None, links=None, embedded=None):
        super(HalExecutor, self).__init__(data, links, embedded)

        self.links.append(link.Link("self", request.path + data["id"]))
        self.links.append(link.Link("update", request.path + data["id"]))
        self.links.append(link.Link("delete", request.path + data["id"]))



from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'VERSION')) as f:
    version = f.read().strip()

with open(path.join(here, 'README.rst')) as f:
    long_description = f.read()

setup(name='chaosmonkey',
      version=version,
      description='A Chaos Engineering swiss army knife',
      author='BBVALabs',
      long_description=long_description,
      url="https://github.com/BBVA/chaos-monkey-engine",
      packages=find_packages(),
      include_package_data=True,
      entry_points='''
        [console_scripts]
        chaos-monkey-engine=chaosmonkey.cm:cm
        ''',
      )



class Attack:
    """
    Base class for attacks. Every attack must extend from this class
    """

    ref = None      #: **string** Unique identifier for the attack. Must be module_name.AttackClass
    schema = None   #: **dict** Valid jsonSchema to validate the attack attributes in the API
    #: **dict** example for using when calling add plan endpoint
    example = None

    def __init__(self, attack_config):
        self.attack_config = attack_config

    def run(self):
        """
        This method is called by an executor to perform the actual attack.
        You can access the self.attack_config that holds the configuration
        used when calling the endpoint to plan the attacks.
        """
        raise NotImplementedError("Attacks should implement this!")

    @staticmethod
    def to_dict():
        """
        You should implement to_dict to return an Attack._to_dict(ref, schema, example)
        using the attack attributes

        Example::

            @staticmethod
            def to_dict():
                return Attack._to_dict(
                    TerminateEC2Instance.ref,
                    TerminateEC2Instance.schema,
                    TerminateEC2Instance.example
                )
        """

        raise NotImplementedError("Attacks should implement this!")

    @staticmethod
    def _to_dict(ref, schema, example):
        return {
            "ref": ref,
            "schema": schema,
            "example": example,
        }



"""
**Base path**: /api/1/plans

Plans receive a planner and an attack and create executors calling the corresponding planner with the given
attack.

Each plan creates N executors related to an attack to be executed in the future.

Plans has the following properties

* **id**: unique identifier for a plan
* **created**: creation date
* **next_execution**: execution date for the next executor
* **name**: plan name
* **executors_count**: number of executors in the plan
* **executed**: if all the executors in the plan has been executed

"""
from flask import Blueprint, json, request
from flask_hal import Document
from chaosmonkey.api.request_validator import validate_payload
from chaosmonkey.engine.cme_manager import manager
from chaosmonkey.api.utils import get_boolean


plans = Blueprint("plans", __name__)

plan_schema = {
    "type": "object",
    "properties": {
        "name": {"type": "string"},
        "attack": {
            "type": "object",
            "properties": {
                "ref": {"type": "string"},
                "args": {"type": "object"},
            },
            "required": ["ref", "args"]
        },
        "planner": {
            "type": "object",
            "properties": {
                "ref": {"type": "string"},
                "args": {"type": "object"},
            },
            "required": ["ref", "args"]
        }
    },
    "required": ["name", "attack", "planner"]
}


@plans.route("/", methods=["POST"])
def add_plan():
    """
    Add a plan.

    Example request::

        PUT /api/1/executors/3b373155577b4d1bbc62216ffea013a4
        Body:
            {
                "name": "Terminate instances in Playground",
                "attack": {
                    "args": {
                        "region": "eu-west-1",
                        "filters": {
                            "tag:Name": "playground-asg"
                        }
                    },
                    "ref": "terminate_ec2_instance:TerminateEC2Instance"
                },
                "planner": {
                    "ref": "simple_planner:SimplePlanner",
                    "args": {
                        "min_time" : "10:00",
                        "max_time" : "19:00",
                        "times": 4
                    }
                }
            }

    """
    assert validate_payload(request, plan_schema)
    req_json = request.get_json()

    name = req_json["name"]
    planner_config = req_json["planner"]
    attack_config = req_json["attack"]

    manager.execute_plan(name, planner_config, attack_config)

    return json.jsonify({"msg": "ok"})


@plans.route("/", methods=["GET"])
def list_plans():
    """
    List all plans created

    Example request::

        GET /api/1/plans/?all=true

    Example response::

        {
            "_links": {
                "self": {
                    "href": "/api/1/plans/"
                }
            },
            "plans": [
                {
                    "id": "6890192d8b6c40e5af16f13aa036c7dc",
                    "created": "2017-01-26T10:41:1485427282",
                    "next_execution": "2017-01-26 13:14:07.583372",
                    "name": "Terminate instances in Playground",
                    "executors_count": 2,
                    "_links": {
                        "self": {
                            "href": "/api/1/plans/6890192d8b6c40e5af16f13aa036c7dc"
                        },
                        "update": {
                            "href": "/api/1/plans/6890192d8b6c40e5af16f13aa036c7dc"
                        },
                        "delete": {
                            "href": "/api/1/plans/6890192d8b6c40e5af16f13aa036c7dc"
                        }
                    }
                }
            ]
        }

    :param: all. Control when to show all plans (true) or only not executed (false). Defaults to false

    :return: :meth:`flask_hal.document`
    """
    show_all_query = request.args.get("all", False)
    show_all = get_boolean(show_all_query)
    plan_list = [plan.to_dict() for plan in manager.get_plans(show_all=show_all)]
    return Document(data={"plans": plan_list})


@plans.route("/<string:plan_id>", methods=["GET"])
def get_plan(plan_id):
    """
    Get a plan with all related executors

    Example request::

        GET /api/1/plans/6890192d8b6c40e5af16f13aa036c7dc

    Example response::

        {
            "id": "6890192d8b6c40e5af16f13aa036c7dc",
            "_embedded": {
                "executors": [
                    {
                        "plan_id": "6890192d8b6c40e5af16f13aa036c7dc",
                        "_links": {
                            "self": {
                                "href": "/api/1/plans/6890192d8b6c40e5af16f13aa036c7dcdd2530572fd04c5aa061f261f82743d3"
                            },
                            "update": {
                                "href": "/api/1/plans/6890192d8b6c40e5af16f13aa036c7dcdd2530572fd04c5aa061f261f82743d3"
                            },
                            "delete": {
                                "href": "/api/1/plans/6890192d8b6c40e5af16f13aa036c7dcdd2530572fd04c5aa061f261f82743d3"
                            }
                        },
                        "next_run_time": "2017-01-26T13:14:1485436447",
                        "id": "dd2530572fd04c5aa061f261f82743d3"
                    },
                    {
                        "plan_id": "6890192d8b6c40e5af16f13aa036c7dc",
                        "_links": {
                            "self": {
                                "href": "/api/1/plans/6890192d8b6c40e5af16f13aa036c7dc1dd3f0d392e545808edb74852213c1ae"
                            },
                            "update": {
                                "href": "/api/1/plans/6890192d8b6c40e5af16f13aa036c7dc1dd3f0d392e545808edb74852213c1ae"
                            },
                            "delete": {
                                "href": "/api/1/plans/6890192d8b6c40e5af16f13aa036c7dc1dd3f0d392e545808edb74852213c1ae"
                            }
                        },
                        "next_run_time": "2017-01-26T18:24:1485455082",
                        "id": "1dd3f0d392e545808edb74852213c1ae"
                    }
                ]
            },
            "created": "2017-01-26T10:41:1485427282",
            "next_execution": null,
            "name": "Terminate instances in Playground",
            "executors_count": null,
            "_links": {
                "self": {
                    "href": "/api/1/plans/6890192d8b6c40e5af16f13aa036c7dc"
                }
            }
        }

    :return: :meth:`flask_hal.document`
    """
    plan = manager.get_plan(plan_id)
    executor_list = [executor.to_dict() for executor in manager.get_executors_for_plan(plan_id)]
    return Document(data=plan.to_dict(), embedded={"executors": executor_list})


@plans.route("/<string:plan_id>", methods=["DELETE"])
def delete_plan(plan_id):
    """
    Delete a plan

    Example request::

        DEL /api/1/plans/6890192d8b6c40e5af16f13aa036c7dc


    """
    manager.delete_plan(plan_id)
    return json.jsonify({
        "msg": "Plan %s successfully deleted" % plan_id
    })
